"Distributed random forest in sklearn with MPI."
import argparse
import string
from typing import Tuple
import time

from mpi4py import MPI
import numpy as np
from sklearn.ensemble import RandomForestClassifier

from dataloaders import load_data_csv_parallel, load_data_csv_root


class MPITimer:
    """A distributed context-manager enabled timer."""

    def __init__(
        self,
        comm: MPI.Comm,
        print_on_exit: bool = True,
        name: str = "",
        output_format: str = "Elapsed time {name}: global average {elapsed_time_average:.2g}s, "
        "local {elapsed_time_local:.2g}s",
    ) -> None:
        """
        Create a new distributed context-manager enabled timer.

        Parameters
        ----------
        comm : MPI.Comm
            The MPI communicator.
        print_on_exit : bool
            Whether to print the measured time in __exit__.
        name : str
            Label describing what this timer measured, can be used for printing the results.
        output_format : str
            Format string template used for printing the output. May reference all attributes of the
        """
        self.comm = comm
        self.output_format = output_format
        self.print_on_exit = print_on_exit
        self.name = name

        self.start_time = None
        self.end_time = None
        self.elapsed_time_local = None
        self.elapsed_time_average = None

    def start(self) -> None:
        """Start the timer by setting the start time."""
        self.start_time = time.perf_counter()

    def stop(self) -> None:
        """Stop the timer by setting the end time and updating elapsed_time_local."""
        self.end_time = time.perf_counter()
        self.elapsed_time_local = self.end_time - self.start_time

    def allreduce_for_average_time(self) -> None:
        """Compute the global average using allreduce and update elapsed_time_average."""
        self.elapsed_time_average = (
            self.comm.allreduce(self.elapsed_time_local, op=MPI.SUM) / self.comm.size
        )

    def print(self) -> None:
        """Print the elapsed time using the given template."""
        template_keywords = {
            key for (_, key, _, _) in string.Formatter().parse(self.output_format)
        }
        template_kwargs = {
            key: value for key, value in vars(self).items() if key in template_keywords
        }
        print(self.output_format.format(**template_kwargs))

    def __enter__(self) -> "MPITimer":
        """
        Called on entering this context (e.g. with a 'with' statement), starts the timer.

        Returns
        -------
        MPITimer
            This timer object.
        """
        self.start()
        return self

    def __exit__(self, *args) -> None:
        """
        Called on exiting this context (e.g. after a 'with' statement), stops the timer, computes the global average
        and optionally prints the result on rank 0.

        Parameters
        ----------
        args :
            Unused, only to fulfill __exit__ interface.
        """
        self.stop()
        self.allreduce_for_average_time()
        if self.print_on_exit and self.comm.rank == 0:
            self.print()


class DistributedRandomForest:
    """
    Distributed random forest class.

    Attributes
    ----------
    acc_global : None | float
        The global accuracy.
    acc_local : None | float
        Each local trained model's accuracy.
    clf : None | RandomForestClassifier
        The sklearn classifier.
    comm : MPI.Comm
        The communicator.
    n_trees_base : int
        The base number of rank-local trees.
    n_trees_global : int
        The number of trees in the global forest.
    n_trees_local : int
        The number of local trees.
    n_trees_remainder : int
        The remaining number of trees to distribute.
    random_state : int
        The base random state for sklearn random forest.

    Methods
    -------
    _distribute_trees()
        Distribute trees evenly over all processors.
    _train_local_classifier()
        Train local random forest classifier.
    _predict_locally()
        Get predictions of all sub estimators in random forest.
    _get_predicted_class_hist()
        Calculate global sample-wise distributions of predicted classes from rank-local tree-wise predictions.
    _calc_majority_vote_hist()
        Calculate majority vote from sample-wise histograms of predicted classes.
    train()
        Train random forest in parallel.
    test()
        Test trained distributed global random forest.
    """

    def __init__(
        self,
        n_trees_global: int,
        comm: MPI.Comm,
        random_state: int,
    ) -> None:
        """
        Initialize distributed random forest object.

        Parameters
        ----------
        n_trees_global : int
            The number of trees in the global forest.
        comm : MPI.Comm
            The communicator to use.
        random_state : int
            The base random state for sklearn RF.
        """
        self.n_trees_global = n_trees_global
        self.comm = comm
        (
            self.n_trees_base,
            self.n_trees_remainder,
            self.n_trees_local,
        ) = self._distribute_trees()
        self.random_state = random_state + self.comm.rank
        self.clf = None  # local random forest classifier
        self.acc_global = None  # accuracy of global classifier
        self.acc_local = None  # accuracy of each rank-local classifier

    def _distribute_trees(self) -> Tuple[int, int, int]:
        """
        Distribute trees evenly over all processors.

        Returns
        -------
        int
            The base number of rank-local trees.
        int
            The remaining number of trees to distribute.
        int
            The final number of rank-local trees.
        """
        size, rank = self.comm.size, self.comm.rank
        n_trees_base = self.n_trees_global // size  # base number of rank-local trees
        n_trees_remainder = (
            self.n_trees_global % size
        )  # remaining number of trees to distribute
        n_trees_local = (
            n_trees_base + 1 if rank < n_trees_remainder else n_trees_base
        )  # final number of local trees
        return n_trees_base, n_trees_remainder, n_trees_local

    def _train_local_classifier(
        self, train_samples: np.ndarray, train_targets: np.ndarray
    ) -> RandomForestClassifier:
        """
        Train local random forest classifier.

        Params
        ------
        train_samples : numpy.ndarray
            samples of train dataset
        train_targets : numpy.ndarray
            targets of train dataset

        Returns
        -------
        sklearn.ensemble.RandomForestClassifier
            trained model
        """
        clf = RandomForestClassifier(
            n_estimators=self.n_trees_local, random_state=self.random_state
        )
        _ = clf.fit(train_samples, train_targets)
        return clf

    def _predict_locally(self, samples: np.ndarray) -> np.ndarray:
        """
        Get predictions of all sub estimators in random forest.

        Parameters
        ----------
        samples : numpy.ndarray
            The samples whose class to predict.

        Returns
        -------
        numpy.ndarray
            The predictions of each sub estimator on the samples.
        """
        return np.array(
            [tree.predict(samples) for tree in self.clf.estimators_], dtype="H"
        )

    def _get_predicted_class_hist(
        self, tree_wise_predictions: np.ndarray, n_classes: int
    ) -> np.ndarray:
        """
        Calculate global sample-wise distributions of predicted classes from rank-local tree-wise predictions.

        Parameters
        ----------
        tree_wise_predictions : numpy.ndarray
            The tree-wise predictions.
        n_classes : int
            The number of classes in the dataset.

        Returns
        -------
        numpy.ndarray
            The sample-wise distributions of the predicted classes.
        """
        sample_wise_predictions = tree_wise_predictions.transpose()
        predicted_class_hists_local = np.array(
            [
                np.bincount(sample_pred, minlength=n_classes)
                for sample_pred in sample_wise_predictions
            ]
        )
        # The hist arrays have `n_test_samples` entries with `n_classes` elements each.
        # The local hist holds the class prediction distribution for each test sample over the local forest.
        # Now we want to sum up those sample-wise distributions to obtain the global hist over the global forest.
        # From this global hist, we can determine the global majority vote for each test sample.
        predicted_class_hists_global = np.zeros_like(predicted_class_hists_local)
        self.comm.Allreduce(predicted_class_hists_local, predicted_class_hists_global)
        return predicted_class_hists_global

    @staticmethod
    def _calc_majority_vote_hist(predicted_class_hists: np.ndarray) -> np.ndarray:
        """
        Calculate majority vote from sample-wise histograms of predicted classes.

        Parameters
        ----------
        predicted_class_hists : numpy.ndarray
            sample-wise histograms of predicted classes

        Returns
        -------
        numpy.ndarray
            majority votes for all samples in histogram input
        """
        return np.array(
            [np.argmax(sample_hist) for sample_hist in predicted_class_hists]
        )

    def train(
        self,
        train_samples: np.ndarray,
        train_targets: np.ndarray,
    ) -> None:
        """
        Train random forest in parallel.

        Parameters
        ----------
        train_samples : numpy.ndarray
            rank-local train samples
        train_targets : numpy.ndarray
            corresponding train targets
        """
        # Set up communicator.
        rank, size = self.comm.rank, self.comm.size
        # Set up and train local forest.
        print(
            f"[{rank}/{size}]: Set up and train local random forest with "
            f"{self.n_trees_local} trees and random state {self.random_state}."
        )
        self.clf = self._train_local_classifier(
            train_samples=train_samples,
            train_targets=train_targets,
        )

    def test(
        self,
        test_samples: np.ndarray,
        test_targets: np.ndarray,
        n_classes: int,
    ) -> None:
        """
        Test trained distributed global random forest.

        Parameters
        ----------
        test_samples : numpy.ndarray
            The rank-local test samples.
        test_targets : numpy.ndarray
            The corresponding test targets.
        n_classes : int
            The number of classes in the dataset.
        """
        rank, size = self.comm.rank, self.comm.size
        # Get class predictions of sub estimators in each forest.
        print(f"[{rank}/{size}]: Get predictions of individual sub estimators.")
        tree_predictions_local = self._predict_locally(test_samples)
        print(f"[{rank}/{size}]: Calculate majority vote via histograms.")
        predicted_class_hists = self._get_predicted_class_hist(
            tree_predictions_local, n_classes
        )
        majority_vote = self._calc_majority_vote_hist(predicted_class_hists)
        # Calculate accuracies.
        self.acc_global = (test_targets == majority_vote).mean()
        self.acc_local = self.clf.score(test_samples, test_targets)
        print(
            f"[{rank}/{size}]: Local accuracy is {self.acc_local}, global accuracy is {self.acc_global}."
        )


if __name__ == "__main__":
    # SETTINGS
    comm = MPI.COMM_WORLD  # Set up communicator.
    rank, size = comm.rank, comm.size

    if rank == 0:
        print(
            "######################################################\n"
            "# Distributed Random Forest in Scikit-Learn with MPI #\n"
            "######################################################\n"
        )

    n_classes = 2

    header_lines = 0  # Data file specifications
    sep = ","
    encoding = "utf-8"
    path_to_data = "/pfs/work7/workspace/scratch/ku4408-VL-ScalableAI/data/SUSY.csv"

    parser = argparse.ArgumentParser(prog="Distributed Random Forest")
    parser.add_argument(
        "--dataloader",
        type=str,
        choices=["root", "parallel"],
        default="root",
        help="The dataloader to use.",
    )
    parser.add_argument(
        "--num_trees",
        type=int,
        default=100,
        help="The global number of trees.",
    )
    parser.add_argument(
        "--train_split",
        type=float,
        default=0.75,
        help="The fraction of the dataset used for training.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="The verbosity.",
    )
    args = parser.parse_args()

    seed = size

    # -------------- Load data. --------------
    print(f"[{rank}/{size}]: Loading data...")
    with MPITimer(comm, name="data loading"):
        if args.dataloader == "parallel":
            if rank == 0:
                print("Using truly parallel dataloader...")
            (
                train_samples_local,
                train_targets_local,
                test_samples,
                test_targets,
            ) = load_data_csv_parallel(
                path_to_data, header_lines, comm, args.train_split, args.verbose
            )

        elif args.dataloader == "root":
            if rank == 0:
                print("Using root-based dataloader with Scatterv...")
            (
                train_samples_local,
                train_targets_local,
                test_samples,
                test_targets,
            ) = load_data_csv_root(
                path_to_data, header_lines, comm, seed, args.train_split, args.verbose
            )
        else:
            raise ValueError

    print(
        f"[{rank}/{size}]: DONE.\n"
        f"Local train samples and targets have shapes {train_samples_local.shape} and {train_targets_local.shape}.\n"
        f"Global test samples and targets have shapes {test_samples.shape} and {test_targets.shape}.\n"
        f"Labels are {train_targets_local}"
    )

    # -------------- Set up random forest. --------------
    with MPITimer(comm, name="forest creation"):
        distributed_random_forest = DistributedRandomForest(
            n_trees_global=args.num_trees,
            comm=comm,
            random_state=seed,
        )

    # -------------- Train random forest. --------------
    with MPITimer(comm, name="training"):
        distributed_random_forest.train(
            train_samples_local,
            train_targets_local,
        )

    # -------------- Evaluate random forest. --------------
    print(f"[{rank}/{size}]: Evaluate random forest.")
    with MPITimer(comm, name="test"):  # Test trained model on test data.
        distributed_random_forest.test(
            test_samples,
            test_targets,
            n_classes,
        )
